"""game.py - a context for running a game session or level
"""

import os

import gsdl2
from gsdl2.locals import Color, S_A, S_D, S_M, S_S, S_W, S_RIGHT, S_LEFT, S_DOWN, S_UP, S_F9, S_F10, S_ESCAPE

import context
import enemy
import player
import settings
import sound
import spatialhash
import weapon


class Game(context.Context):
    
    def __init__(self):
        context.Context.__init__(self, gsdl2.time.FixedDriver(self.update, settings.time_step))
        self.render_sched = self.clock.new_schedule(self.draw, 0.0, keep_history=True)
        self.renderer = gsdl2.display.get_renderer()
        font_name, font_size = settings.font_info['score']
        self.font = gsdl2.font.Font(os.path.join(settings.font_dir, font_name), font_size)
        self.font_color = Color('green')

        self.space = spatialhash.SpatialHash(gsdl2.Rect((0, 0), settings.resolution), 32)

        self.score_tex = None
        self.score_rect = None

        self.player = player.Player()
        self.space.add(self.player)
        self.new_enemy_bullets = None
        self.new_player_bullets = None
        self.movex = 0
        self.movey = 0

        self.enemy_reset = settings.enemy_rate
        self.enemy_timer = self.enemy_reset
        self.clock.new_schedule(self.set_caption, 1.0)

        self.num_spawn = 0
        self.bonus_spawn = 0
        self.enemies_hit = 0
        self.enemies_total = 0
        self.player_hit = 0
        self.gain_weapon = False
        self.lose_weapon = False

        self.max_objects = 0
        self.load_test = False

        if settings.play_music:
            sound.start_music()
        gsdl2.music.set_endevent(code=settings.SONG_FINISHED)

    def set_caption(self, dt):
        try:
            num_objects = len(self.space)
            self.max_objects = max(self.max_objects, num_objects)
            gsdl2.display.set_caption(
                '{} fps | Enemy rate: {:0.3f} | Hit ratio: {}% ({}/{}) | Objects: {} (max={}) | Num spawn: {}'.format(
                    int(self.render_sched.per_second()),
                    self.enemy_reset,
                    int(float(self.enemies_hit) / self.enemies_total * 100),
                    self.enemies_hit, self.enemies_total,
                    num_objects,
                    self.max_objects,
                    self.num_spawn))
        except ZeroDivisionError:
            pass

    def update(self, dt):
        self.get_events()
        self.update_move_player()
        self.update_world(dt)
        self.update_collisions()
        self.update_add_bullets()
        self.update_new_enemy(dt)
        self.update_player_weapons()
        self.update_score()

    def update_move_player(self):
        """update the player object"""
        if self.movex or self.movey:
            self.player.move_pos(self.movex, self.movey)
            self.space.add(self.player)

    def update_world(self, dt):
        """update objects in space, fire bullets, remove derelict objects"""
        Player = player.Player
        Enemy = enemy.Enemy
        space_add = self.space.add
        space_remove = self.space.remove
        colliderect = settings.playfield_rect.colliderect
        new_enemy_bullets = []
        new_player_bullets = []
        enemy_bullets_extend = new_enemy_bullets.extend
        for obj in list(self.space):
            # fire bullets
            try:
                # Object counts will skyrocket if the player allows many enemies and bullets to build up.
                # This could happen if the player has a bad wave that takes out many of his weapons. In
                # this case we'll curtail the bullets until the player recovers sufficiently. This avoids
                # (hopefully) a CPU spiral of death.
                if isinstance(obj, Enemy):
                    if len(self.space) < settings.max_objects_allowed:
                        enemy_bullets_extend(obj.weapon.fire())
                elif isinstance(obj, Player):
                    new_player_bullets.extend(obj.weapon.fire())
            except AttributeError:
                pass
            # update objects
            obj.update(dt)
            try:
                space_add(obj)
            except IndexError:
                pass
            # remove derelict objects
            if not colliderect(obj.rect):
                space_remove(obj)
        self.new_enemy_bullets = new_enemy_bullets
        self.new_player_bullets = new_player_bullets
        # play sounds
        if new_player_bullets:
            sound.play_player_shoot()
        num_enemy_bullets = len(new_enemy_bullets) / 15
        num_enemy = len([o for o in self.space if isinstance(o, Enemy)])
        if num_enemy_bullets > num_enemy / 4:
            sound.play_enemy_shoot()

    def update_collisions(self):
        """handle collisions:
        friendly bullet vs enemy
        enemy bullet vs player
        friendly bullet vs enemy bullet
        """
        Player = player.Player
        Enemy = enemy.Enemy
        FriendlyBullet = weapon.FriendlyBullet
        EnemyBullet = weapon.EnemyBullet
        collisions = self.space.collidealllist()
        enemies = []
        for obj, other in collisions:
            if isinstance(obj, FriendlyBullet) and isinstance(other, Enemy):
                self.space.remove(obj)
                self.space.remove(other)
                enemies.append(other)
                self.gain_weapon = True
            elif isinstance(obj, EnemyBullet) and isinstance(other, Player):
                self.space.remove(obj)
                self.player_hit += 1
                self.lose_weapon = True
            elif isinstance(obj, FriendlyBullet) and isinstance(other, EnemyBullet):
                self.space.remove(obj)
                self.space.remove(other)
        self.enemies_hit += len(set(enemies))

    def update_add_bullets(self):
        num_bullets = len(self.new_enemy_bullets)
        num_space = len(self.space)
        new_total = num_space + num_bullets
        try:
            self.space.addlist(self.new_player_bullets)
            if new_total > settings.max_objects_allowed:
                new_total = settings.max_objects_allowed - num_space
                self.space.addlist(self.new_enemy_bullets[:new_total])
            else:
                self.space.addlist(self.new_enemy_bullets)
        except IndexError:
            pass

    def update_player_weapons(self):
        # Gain a weapon
        if self.enemies_hit % settings.reward_rate == 0:
            if self.gain_weapon:
                if self.player.gain_weapon():
                    sound.play_gain_weapon()
                self.gain_weapon = False
        if self.player_hit % settings.penalty_rate == 0:
            if self.lose_weapon:
                if self.player.lose_weapon():
                    sound.play_lose_weapon()
                self.lose_weapon = False

    def update_new_enemy(self, dt):
        """Create a new enemy on enemy_timer"""
        self.enemy_timer -= dt
        if self.enemy_timer <= 0.0:
            # if not self.load_test:
            ap = self.player.weapon.active_ports
            np = self.player.weapon.num_ports
            difficulty = float(ap) / np
            self.num_spawn = 2 * ap + int(difficulty * difficulty * self.bonus_spawn)
            #
            self.space.addlist(enemy.enemy_factory(self.num_spawn))
            self.enemies_total += self.num_spawn
            self.enemy_timer += self.enemy_reset
            if self.player.weapon.active_ports == self.player.weapon.num_ports and \
                    len(self.space) < settings.max_objects_allowed:
                self.bonus_spawn += settings.end_game_bonus_per_wave

    def update_score(self):
        score_surf = self.font.render('{}'.format(self.enemies_hit), self.font_color)
        self.score_tex = gsdl2.texture.Texture(self.renderer, score_surf)
        self.score_rect = self.score_tex.get_rect(topleft=(10, 10))

    def draw(self, sched):
        interp = sched.interp
        self.renderer.clear()
        for obj in self.space:
            self.renderer.copy(obj.texture, obj.draw_rect(interp))
        if self.score_tex:
            self.renderer.copy(self.score_tex, self.score_rect)
        self.renderer.present()

    def on_key_down(self, e):
        if e.scancode in (S_D, S_RIGHT):
            self.movex += self.player.speed
        elif e.scancode in (S_S, S_DOWN):
            self.movey += self.player.speed
        elif e.scancode in (S_A, S_LEFT):
            self.movex -= self.player.speed
        elif e.scancode in (S_W, S_UP):
            self.movey -= self.player.speed
        elif e.scancode == S_M:
            if settings.play_music:
                settings.play_music = False
                sound.stop_music()
            else:
                settings.play_music = True
                sound.start_music()
        elif e.scancode == S_F9:
            # CHEAT: max active ports
            weapon = self.player.weapon
            while weapon.active_ports < weapon.num_ports:
                weapon.gain_port()
            self.num_spawn = 100
        elif e.scancode == S_F10:
            # CHEAT: for testing, force bonus spawn and max objects increase
            settings.max_objects_allowed += 200
            self.bonus_spawn += settings.end_game_bonus_per_wave
        elif e.scancode == S_ESCAPE:
            self.on_quit()

    def on_key_up(self, e):
        if e.scancode in (S_D, S_RIGHT):
            self.movex -= self.player.speed
        elif e.scancode in (S_S, S_DOWN):
            self.movey -= self.player.speed
        elif e.scancode in (S_A, S_LEFT):
            self.movex += self.player.speed
        elif e.scancode in (S_W, S_UP):
            self.movey += self.player.speed

    def on_user_event(self, e):
        if e.code == settings.SONG_FINISHED:
            if settings.play_music:
                sound.queue_music()

    def on_quit(self):
        gsdl2.music.set_endevent(None)
        sound.stop_music()
        self.pop()
