import math
import random

import gsdl2

import settings
import geometry


WEAPON_ANGLES = (
    0.0,                    # 0
    15.0, 30.0, 45.0,       # 1-3
    60.0, 75.0, 90.0,       # 4-6
    105.0, 120.0, 135.0,    # 7-9
    150.0, 165.0, 180.0,    # 10-12
    195.0, 210.0, 225.0,    # 13-15
    240.0, 255.0, 270.0,    # 16-18
    285.0, 300.0, 315.0,    # 19-21
    330.0, 345.0,           # 22-23
)


class _Bullet(object):

    size = None
    color = None
    texture = None
    speed = None

    def __init__(self, pos, angle):
        self.x, self.y = pos
        self.angle = angle
        if self.__class__.texture is None:
            w, h = settings.resolution
            w = int(math.ceil(max(w / 200, h / 200)))
            _Bullet.size = w, w
            image = gsdl2.Surface((w, w))
            image.fill(gsdl2.Color(self.color))
            self.__class__.texture = gsdl2.Texture(gsdl2.display.get_renderer(), image)
        self.rect = self.texture.get_rect(center=pos)

    def draw_rect(self, interp):
        return self.rect

    def update(self, dt):
        x = self.x
        y = self.y
        x, y = geometry.point_on_circumference((x, y), self.speed, self.angle)
        self.rect.center = x, y
        self.x = x
        self.y = y


class FriendlyBullet(_Bullet):
    texture = None
    color = 'gold'
    speed = int(math.ceil(settings.resolution[0] // 50))


class EnemyBullet(_Bullet):
    texture = None
    color = 'darkviolet'
    speed = int(math.ceil(settings.resolution[0] // (150, 200)[1]))


class _Weapon(object):

    size = None
    BulletType = None
    reset_group = 1.0
    reset_shot = 1.0
    weapon_angles = WEAPON_ANGLES

    def __init__(self, ship, sites=None, shots_per_group=1):
        self.ship = ship
        self.shots_per_group = shots_per_group
        if _Weapon.size is None:
            w = settings.resolution[0] // 400
            _Weapon.size = w, w
        self.rect = ship.rect
        self.ready = False
        self.shot_interval = self.reset_shot / (self.shots_per_group * 2)
        self.cool_down_shot = random.randrange(shots_per_group) * 3 * self.shot_interval
        self.cool_down_group = self.cool_down_shot
        self.shots = [self.shot_interval] * self.shots_per_group
        self.num_ports = len(self.weapon_angles)
        self.weapon_sites = [0] * self.num_ports
        for s in sites:
            self.weapon_sites[s] = True
        self.weapon_indices = range(int(round(self.num_ports // 2)))
        self.active_ports = len([s for s in self.weapon_sites if s])

    def update(self, dt):
        self.cool_down_group -= dt
        if self.cool_down_group <= 0.0:
            self.cool_down_group += self.reset_group
            # self.ready = True
            self.cool_down_shot = 0.0
            self.shots = [self.shot_interval] * self.shots_per_group
        self.cool_down_shot -= dt
        if self.cool_down_shot <= 0.0 and self.shots:
            self.cool_down_shot += self.shots.pop()
            self.ready = True

    def fire(self):
        bullets = []
        if not self.ready:
            return bullets
        for i, s in enumerate(self.weapon_sites):
            if s:
                bullets.append(self.BulletType(self.ship.rect.center, self.weapon_angles[i]))
        self.ready = False
        return bullets

    def gain_port(self):
        if self.active_ports < self.num_ports:
            weapon_sites = self.weapon_sites
            last = len(weapon_sites) - 1
            for i in self.weapon_indices:
                if not weapon_sites[i]:
                    weapon_sites[i] = True
                    self.active_ports += 1
                    return True
                elif not weapon_sites[last - i]:
                    weapon_sites[last - i] = True
                    self.active_ports += 1
                    return True
        return False

    def lose_port(self):
        if self.active_ports > 1:
            weapon_sites = self.weapon_sites
            last = len(weapon_sites) - 1
            for i in reversed(self.weapon_indices):
                if weapon_sites[i]:
                    weapon_sites[i] = False
                    self.active_ports -= 1
                    return True
                elif weapon_sites[last - i]:
                    weapon_sites[last - i] = False
                    self.active_ports -= 1
                    return True
        return False


class FriendlyWeapon(_Weapon):
    BulletType = FriendlyBullet
    reset_group = settings.player_fire_rate
    weapon_angles = [float(i) for i in range(0, 360, 6)]


class EnemyWeapon(_Weapon):
    BulletType = EnemyBullet
    reset_group = settings.enemy_fire_rate
    reset_shot = 3.0
